# coding=utf-8
import mod.server.extraServerApi as serverApi
from mod.common.utils.mcmath import Vector3

CustomGoalCls = serverApi.GetCustomGoalCls()
CompFactory = serverApi.GetEngineCompFactory()


class DestroyWall(CustomGoalCls):
    def __init__(self, entityId, argsJson):
        super(DestroyWall, self).__init__(entityId, argsJson)
        self.mEntityId = entityId
        self.mTimeCounter = 0
        self.mTargetBlockDestroyTime = 0  # 目标方块破坏时间

    # region 继承函数
    def CanUse(self):
        if self._HasTarget() and self._IsTargetAlive() and self._HasBlockBetweenTargetAndCanDestroyBlock():
            return True
        return False

    def CanContinueToUse(self):
        return self.CanUse()

    def CanBeInterrupted(self):
        return True

    def Start(self):
        self.mTimeCounter = 0
        self._FaceToTarget()
        self._ResetDestroyTime()  # 开始执行时设置目标方块破坏时间
        self._TriggerEntityEvent("tutorial:destroy_block_attack")  # 设置超远距离的攻击

    def Stop(self):
        self._TriggerEntityEvent("tutorial:cancel_destroy_block_attack")  # 还原普通攻击

    def Tick(self):
        self.mTimeCounter += 1
        time2destroy = self.mTimeCounter >= int(self.mTargetBlockDestroyTime * 20)  # 1 秒 20 帧
        perSec = self.mTimeCounter % 20 == 0
        if time2destroy:
            self._DestroyFrontBlock()
            self._ResetDestroyTime()
        elif perSec:
            # 使用 /say 命令来监控实体当前的状态
            leftTime = self.mTargetBlockDestroyTime - self.mTimeCounter / 20.0
            CompFactory.CreateCommand(self.mEntityId).SetCommand('/say 正在破坏方块，还剩:{}s'.format(leftTime))

    # endregion

    # region 类函数
    def _HasTarget(self):
        #是否有仇恨目标
        comp = CompFactory.CreateAction(self.mEntityId)
        targetId = comp.GetAttackTarget()
        hasTarget = targetId != "-1"
        return hasTarget

    def _IsTargetAlive(self):
        comp = CompFactory.CreateAction(self.mEntityId)
        targetId = comp.GetAttackTarget()
        comp = CompFactory.CreateGame(serverApi.GetLevelId())
        alive = comp.IsEntityAlive(targetId)
        return alive

    # 跟目标之间是否存在方块并且是否能够破坏
    def _HasBlockBetweenTargetAndCanDestroyBlock(self):
        blockDictList = self._GetFirstBlockFromRay()
        if len(blockDictList) == 0:
            return False
        blockDict = blockDictList[0]
        blockPos = blockDict['pos']
        selfPos = CompFactory.CreatePos(self.mEntityId).GetFootPos()
        distance2block = (Vector3(blockPos) - Vector3(selfPos)).Length()
        if distance2block > 2:
            # 2 是手长，也就是只有够得到的方块才能够被破坏
            return False
        # 还需要检测手上的物品是否能够破坏掉方块
        destroyTotalTime = self._GetTargetBlockDestroyTime(blockDict['identifier'])
        canDestroy = destroyTotalTime > 0
        return canDestroy

    def _GetTargetBlockDestroyTime(self, blockIdentifier):
        blockInfoComp = CompFactory.CreateBlockInfo(self.mEntityId)
        carriedItem = CompFactory.CreateItem(self.mEntityId).GetEntityItem(serverApi.GetMinecraftEnum().ItemPosType.CARRIED, 0)
        totalTime = blockInfoComp.GetDestroyTotalTime(blockIdentifier, None if not carriedItem else carriedItem['itemName'])
        return totalTime

    # 面向目标
    def _FaceToTarget(self):
        comp = CompFactory.CreateAction(self.mEntityId)
        targetId = comp.GetAttackTarget()
        targetPosX, targetPosY, targetPosZ = CompFactory.CreatePos(targetId).GetFootPos()
        entityPosX, entityPosY, entityPosZ = CompFactory.CreatePos(self.mEntityId).GetFootPos()
        diffPos = (targetPosX - entityPosX, targetPosY - entityPosY, targetPosZ - entityPosZ)
        CompFactory.CreateRot(self.mEntityId).SetRot(serverApi.GetRotFromDir(diffPos))

    # 破坏生物面前的方块
    def _DestroyFrontBlock(self):
        blockDictList = self._GetFirstBlockFromRay()
        if blockDictList:
            blockDict = blockDictList[0]
            blockPos = blockDict['pos']
            blockInfoComp = CompFactory.CreateBlockInfo(self.mEntityId)
            dimensionId = CompFactory.CreateDimension(self.mEntityId).GetEntityDimensionId()
            blockInfoComp.SetBlockNew(blockPos, {'name': 'minecraft:air'}, 0, dimensionId)

    def _ResetDestroyTime(self):
        blockDict = self._GetFirstBlockFromRay()[0]
        self.mTargetBlockDestroyTime = self._GetTargetBlockDestroyTime(blockDict['identifier'])
        # 使用 /say 命令在游戏中输出破坏时间
        CompFactory.CreateCommand(self.mEntityId).SetCommand('/say 破坏时间：{}s'.format(self.mTargetBlockDestroyTime))

    # 获得视线范围内第一个方块信息
    def _GetFirstBlockFromRay(self):
        dimensionId = CompFactory.CreateDimension(self.mEntityId).GetEntityDimensionId()
        comp = CompFactory.CreateAction(self.mEntityId)
        targetId = comp.GetAttackTarget()
        targetPosX, targetPosY, targetPosZ = CompFactory.CreatePos(targetId).GetFootPos()
        entityPosX, entityPosY, entityPosZ = CompFactory.CreatePos(self.mEntityId).GetFootPos()

        # 这里获取碰撞箱的高度，因为 FootPos 是从脚底开始计算的坐标，这里加上碰撞箱 -0.2 模拟的是头部眼睛的高度
        _, collisionBoxHeight = CompFactory.CreateCollisionBox(self.mEntityId).GetSize()
        _, targetCollisionBoxHeight = CompFactory.CreateCollisionBox(targetId).GetSize()

        targetPosY = targetPosY + targetCollisionBoxHeight - 0.2
        entityPosY = entityPosY + collisionBoxHeight - 0.2

        rot = (targetPosX - entityPosX, targetPosY - entityPosY, targetPosZ - entityPosZ)
        distance = int(Vector3(rot).Length() - 1)   # 这个地方用小数会非常卡，所以取一个 int，主要是防止射线击穿实体

        blockDictList = serverApi.getEntitiesOrBlockFromRay(
            dimensionId, (entityPosX, entityPosY, entityPosZ), rot, distance, False,
            serverApi.GetMinecraftEnum().RayFilterType.OnlyBlocks
        )

        return blockDictList

    def _TriggerEntityEvent(self, eventName):
        comp = CompFactory.CreateEntityEvent(self.mEntityId)
        comp.TriggerCustomEvent(self.mEntityId, eventName)
        # 使用 /say 命令输出打印
        CompFactory.CreateCommand(self.mEntityId).SetCommand('/say 触发自定义事件：{}'.format(eventName))

    # endregion
